//==================================================
// FILE:	digidoc.c
// PROJECT:     Digi Doc
// DESCRIPTION: Utility program to demonstrate the
//   functionality of DigiDocLib
// AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
//==================================================
// Copyright (C) AS Sertifitseerimiskeskus
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// GNU Lesser General Public Licence is available at
// http://www.gnu.org/copyleft/lesser.html
//==========< HISTORY >=============================
//      12.01.2004      Veiko Sinivee
//                      Creation
//==================================================

#include "DigiDocLib.h"
#include "DigiDocConfig.h"
#include "DigiDocPKCS11.h"
#include "DigiDocSAXParser.h"
#include "DigiDocEncSAXParser.h"
#include "DigiDocEnc.h"
#include "DigiDocEncGen.h"
#include "DigiDocConvert.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/objects.h>
#include <sys/stat.h>
#include <time.h>

//==========< global constants >====================

char g_szProgName[] = "DigiDoc";
char g_szProgVer[] = "1.97";


// programm arguments
char* p_szInFile = 0;
char* p_szInEncFile = 0;
char* p_szInDecFile = 0;
char* p_szOutFile = 0;
char* p_szConfigFile = 0;
int p_parseMode = 1;

//--------------------------------------------------
// reads in a signed doc
//--------------------------------------------------
int cmdReadDigiDoc(SignedDoc** ppSigDoc, int nMode)
{
  int err = ERR_OK;
  int nMaxDfLen;
  time_t t1, t2;
  // test memory interface
  /*char* p;
  FILE* f;
  struct stat st;*/

  if(!p_szInFile) {
    printf("No input file specified!\n");
    return -1;
  }
  nMaxDfLen = ConfigItem_lookup_int("DATAFILE_MAX_CACHE", 0);
  
  /*stat(p_szInFile, &st);
  if((f = fopen(p_szInFile, "rb")) != NULL) {
    p = malloc(st.st_size);
    memset(p, 0, st.st_size);
    fread(p, st.st_size, 1, f);
    fclose(f);
    err = ddocSaxReadSignedDocFromMemory(ppSigDoc, p, st.st_size, nMaxDfLen);
    free(p);
    }*/
  time(&t1);
  switch(nMode) {
  case 2: // new XMLReader interface
    //err = ddocXRdrReadSignedDocFromFile(p_szInFile, CHARSET_ISO_8859_1, ppSigDoc, nMaxDfLen);
    break;
  case 1:
  default:
    err = ddocSaxReadSignedDocFromFile(ppSigDoc, p_szInFile, 0, nMaxDfLen);
  }
  time(&t2);
  if(err != ERR_OK)  {
    printf("Error reading file: %s - %d\n", p_szInFile, err);
    SET_LAST_ERROR(err);
  }
  printf("Read document: %d, time %d [sek]\n", err, t2-t1);
  //else
  //err = checkDigiDocErrors();
  return err;
}

//--------------------------------------------------
// writes a digidoc in file
//--------------------------------------------------
int cmdWrite(SignedDoc* pSigDoc, DEncEncryptedData* pEncData)
{
  int err = ERR_OK;
  FILE* hFile;
  
  if(!p_szOutFile) {
    printf("No output file specified!\n");
    return -1;
  }
  if(pSigDoc)
    err = createSignedDoc(pSigDoc, p_szInFile, p_szOutFile);
  if(pEncData) {
    if(pEncData->nDataStatus == DENC_DATA_STATUS_ENCRYPTED_AND_NOT_COMPRESSED ||
       pEncData->nDataStatus == DENC_DATA_STATUS_ENCRYPTED_AND_COMPRESSED) {
      err = dencGenEncryptedData_writeToFile(pEncData, p_szOutFile);
    } else {
      if((hFile = fopen(p_szOutFile, "wb")) != NULL) {
	fwrite((pEncData)->mbufEncryptedData.pMem, 1, 
	       (pEncData)->mbufEncryptedData.nLen, hFile);
	fclose(hFile);
      }
      else
	err = ERR_FILE_WRITE;
    }
  }
  if (err)  {
   	printf("Error writing file: %s - %d\n", p_szOutFile, err);
	SET_LAST_ERROR(err);   
  }
  //else
  //  err = checkDigiDocErrors();
  return err;
}

//--------------------------------------------------
// Creates a new signed doc
//--------------------------------------------------
int cmdNew(SignedDoc** ppSigDoc, const char* pFormat, const char* pVersion)
{
  int err = ERR_OK;
  const char* format = pFormat, *version = pVersion;

  if(!p_szOutFile) {
    printf("No output file specified!\n");
    return -1;
  }
  if(!format)
    format = ConfigItem_lookup("DIGIDOC_FORMAT");
  if(!version)
    version = ConfigItem_lookup("DIGIDOC_VERSION");
  if(format && version) {
     err = SignedDoc_new(ppSigDoc, format, version);
     RETURN_IF_NOT(err == ERR_OK, err);
//     err = checkDigiDocErrors();
  } else {
    printf("Error finding new document format or version!\n");
    SET_LAST_ERROR_RETURN_CODE(ERR_UNSUPPORTED_FORMAT);
  }
  return err;
}

//--------------------------------------------------
// Converts input data to UTF-8
// src - input data
// returns converted string or NULL. Caller must free it.
//--------------------------------------------------
int convertInput(const char* src, char** dest)
{
  int l1, err = ERR_OK;
  char* p = NULL;

  *dest = 0;
  l1 = strlen(src) * 2;
  *dest = (char*)malloc(l1);
  if(*dest) {
    memset(*dest, 0, l1);
#ifdef WIN32
    err = oem2utf8(src, *dest, l1);
#else
    ascii2utf8(src, *dest, &l1);
#endif
  }
  return err;
}
//--------------------------------------------------
// Handles one argument
//--------------------------------------------------
int checkArguments(int argc, char** argv, int* pCounter, char** dest)
{
  int nLen = 0;

  if(argc > (*pCounter)+1 && argv[(*pCounter)+1][0] != '-') {
    nLen = strlen(argv[(*pCounter)+1]) * 2;
    convertInput((const char*)argv[(*pCounter)+1], dest);
    (*pCounter)++;
  }
  return nLen;
}

//--------------------------------------------------
// Adds a DataFile to signed doc
//--------------------------------------------------
int cmdAddDataFile(SignedDoc** ppSigDoc, const char* file, 
		   const char* mime, const char* content, const char* charset)
{
  int err = ERR_OK;
  DataFile  *pDataFile;
  char *p = 0;

  // if there was no new command then implicitly create a new document
  if(!(*ppSigDoc)) {
    err = cmdNew(ppSigDoc, NULL, NULL);
    RETURN_IF_NOT(err == ERR_OK, err); 
  }
  // convert to UTF-8
  err = convertInput(file, &p);
  // add a file	
  err = DataFile_new(&pDataFile, *ppSigDoc, NULL, p, 
		     content, mime, 0, NULL, 0, NULL, charset);
  if(!err)
    err = calculateDataFileSizeAndDigest(*ppSigDoc, pDataFile->szId, p, DIGEST_SHA1);
  if(p)
    free(p);
  RETURN_IF_NOT(err == ERR_OK, err);
  // now calculate file length and digest
  RETURN_IF_NOT(err == ERR_OK, err)
    return err;
}



//--------------------------------------------------
// Verifys sigantures and notaries
//--------------------------------------------------
int cmdVerify(SignedDoc* pSigDoc)
{
  int err = ERR_OK, s, d;
  SignatureInfo* pSigInfo;
  char buf1[100];

  d = getCountOfSignatures(pSigDoc);
  for(s = 0; s < d; s++) {
    pSigInfo = getSignature(pSigDoc, s);
    memset(buf1, 0, sizeof(buf1));
    getSignerLastName(pSigInfo, buf1);
    strcat(buf1, ",");
    getSignerFirstName(pSigInfo, strchr(buf1, 0));
    strcat(buf1, ",");
    getSignerCode(pSigInfo, strchr(buf1, 0));
    printf("Signature: %s - %s\n", pSigInfo->szId, buf1);
    err = verifySignatureAndNotary(pSigDoc, pSigInfo, p_szInFile);
    printf("\tSignature - %s\n", ((!err) ? "OK" : "ERROR"));
    RETURN_IF_NOT(err == ERR_OK, err);
  }
  return err;
}


//--------------------------------------------------
// Signs the document and gets configrmation
//--------------------------------------------------
int cmdSign(SignedDoc* pSigDoc, const char* pin, const char* manifest,
			const char* city, const char* state, const char* zip, const char* country)
{
  int err = ERR_OK;
  SignatureInfo* pSigInfo = NULL;
	
  printf("Creating new digital signature\n");
  err = signDocument(pSigDoc, &pSigInfo, pin, manifest, city, state, zip, country);
  RETURN_IF_NOT(err == ERR_OK, err);
  return ERR_OK;
}

/*void printExtensionsByOid(X509* pCert, const char* oid)
{
  int nid, pos, i;
  X509_EXTENSION* pExt;
  
  nid = OBJ_txt2nid(oid);
  printf("OID: %s -> NID: %d\n", oid, nid);
  pos = X509_get_ext_by_NID(pCert, nid, -1);
  if(pos >= 0) {
    pExt = X509_get_ext(pCert, pos);
    printf("OID: %s len: %d data: \'%s\'\n", oid, 
	   ((ASN1_STRING*)pExt->value)->length,
	   ((ASN1_STRING*)pExt->value)->data);
  }
  else
    printf("\tnot found!\n");
}

void cmdExtensions(const char* szCertFile)
{
  X509* pCert;
  PolicyIdentifier* pPolicies;
  int nPols, err, i;
  char buf[500];

  err = ReadCertificate(&pCert, szCertFile);
  if(err) checkDigiDocErrors();
  err = ReadCertSerialNumber(buf, pCert);
  if(err) checkDigiDocErrors();
  printf("Certificate No: %s\n", buf);
  
  err = readCertPolicies(pCert, &pPolicies, &nPols);
  if(err) checkDigiDocErrors();
  for(i = 0; i < nPols; i++) {
    printf("Policy OID: %s URL: %s Description: %s\n", 
	   pPolicies[i].szOID, pPolicies[i].szCPS, pPolicies[i].szUserNotice);
  }
  PolicyIdentifiers_free(pPolicies, nPols);
  // debug loog for these oid-s
  printExtensionsByOid(pCert, "1.3.6.1.5.5.7.1.3");
  printExtensionsByOid(pCert, "2.5.29.9");
  printExtensionsByOid(pCert, "1.3.6.1.5.5.7.1.2");

  X509_free(pCert);
  }*/



//--------------------------------------------------
// prints usage statement
//--------------------------------------------------
void printUsage()
{
	printf("USAGE: digidoc <command(s)> [-in <input-file>] [-out <output-file>] [-config <config-file>]\n");
	printf("COMMANDS:\n");
	printf("\t[-new] [format] [version]\n");
	printf("\t-add <file-name> <mime-type> [<content-type>] [<charset>]\n");
	printf("\t-verify \n");
	printf("\t-sign <pin> [<manifest>] [<city> <state> <zip> <country>]\n");
	printf("\t-extract <doc-id> <file-name> [<charset>] [<file-name-charset>]\n");
	printf("\t-encrypt <file-name>\n");
	printf("\t-encrecv <cert-file> [<recipient>] [<keyname>] [<carried-key-name>]\n");
	printf("\t-decrypt <file-name> <recipient> <pin1>\n");
	printf("\t-denc-list <file-name>\n");
	printf("\t-encrypt-file <in-file-name> <out-file-name> [<in-file-mime-type>]\n");
	printf("\t-decrypt-file <in-file-name> <out-file-name> <recipient-cert-cn> <pin1>\n");
	//printf("\t-extensions <certificate-in-pem-format>\n");
	printf("\t-test <digidoc-filename>\n");
}

void cmdTest()
{

}

//--------------------------------------------------
// Encrypts a file
//--------------------------------------------------
int cmdEncrypt(DEncEncryptedData** ppEncData, const char* szFileName) 
{
  FILE* hFile;
  DEncEncryptionProperty* pEncProperty = 0;
  int len, err = ERR_OK;
  char buf[2048], *p;

  err = dencEncryptedData_new(ppEncData, DENC_XMLNS_XMLENC, DENC_ENC_METHOD_AES128, 0, 0, 0);
  if(err) return err;
  p = 0;
  err = convertInput(szFileName, &p);
  if(err) return err;
  err = dencEncryptionProperty_new(*ppEncData, &pEncProperty, 0, 0, ENCPROP_FILENAME, getSimpleFileName(p));
  free(p);
  if(err) return err;
  if((hFile = fopen(szFileName, "rb")) != NULL) {
    do {
      memset(buf,0,sizeof(buf));
      len = fread(buf, 1, sizeof(buf), hFile);
      if(len)
	err = dencEncryptedData_AppendData(*ppEncData, buf, len);
    } while(len && !err);
    fclose(hFile);
  }
  if(!err)
    err = dencEncryptedData_encryptData(*ppEncData, DENC_COMPRESS_NEVER); //DENC_COMPRESS_BEST_EFFORT);
  return err;
}


//--------------------------------------------------
// Checks program runtime arguments
//--------------------------------------------------
int checkProgArguments(int argc, char** argv)
{
  int i;
  int err = ERR_OK;
  for(i = 1; !err && (i < argc); i++) {
    if(!strcmp(argv[i], "-?") || !strcmp(argv[i], "-help")) {
      printUsage();
      continue;
    }
    //printf("ARG: %d - %s\n", i, argv[i]);
    if (argv[i][0] == '-') { // all commands and options must start with -
      if (!strcmp(argv[i], "-in")) {
	//if(!checkArguments(argc, argv, &i, &p_szInFile)))
	if(argc > i+1 && argv[i+1][0] != '-') {
	  p_szInFile = argv[i+1];
	  i++;
	} else {
	  printf("Missing or invalid input file name\n");
	  err = -1;
	}
	if(argc > i+1 && argv[i+1][0] != '-') {
	  if(!strcmp(argv[i+1], "SAX"))
	    p_parseMode = 1;
	  if(!strcmp(argv[i+1], "XRDR"))
	    p_parseMode = 2;
	  i++;
	}
      }
      if (!strcmp(argv[i], "-out")) {
	if (argc > i+1 && argv[i+1][0] != '-') {
	  p_szOutFile = argv[i+1];
	  i++;
	} else {
	  printf("Missing or invalid output file name\n");
	  err = -1;
	}
      }
      if (!strcmp(argv[i], "-encrypt")) {
	if (argc > i+1 && argv[i+1][0] != '-') {
	  p_szInEncFile = argv[i+1];
	  i++;
	} else {
	  printf("Missing or invalid encrypt input file name\n");
	  err = -1;
	}
      }
      if (!strcmp(argv[i], "-config")) {
	if (argc > i+1 && argv[i+1][0] != '-') {
	  p_szConfigFile = argv[i+1];
	  i++;
	} else {
	  printf("Missing or invalid configuration file name\n");
	  err = -1;
	}
      }
      /*if (!strcmp(argv[i], "-extension")) {
	if (argc > i+1 && argv[i+1][0] != '-') {
	cmdExtensions(argv[i+1]);
	exit(0);
	} else {
	printf("Missing or invalid certificate file name\n");
	err = -1;
	}
	}*/
      if (!strcmp(argv[i], "-test")) {
	cmdTest();
	exit(0);
      }
    } /*else	 {
	printf("Invalid option: %s that does not start with -\n", argv[i]);
	err = -1; // TODO: accept first option as filename and use default command verify
	}*/
  }
  if(err)
    printUsage();
  return err;
}

//--------------------------------------------------
// Create digidoc and adds datafiles
//--------------------------------------------------
int runAddCmds(int argc, char** argv, SignedDoc** ppSigDoc)
{
	int err = ERR_OK, i;

	for(i = 1; (err == ERR_OK) && (i < argc); i++) {
	  if(argv[i][0] == '-') { // all commands and options must start with -
	    // create new digidoc
	    if(!strcmp(argv[i], "-new")) {
	      char* format = NULL;
	      char* version = NULL;
	      // optional content and charset
	      if(argc > i+1 && argv[i+1][0] != '-') {
		format = argv[i+1];
		i++;
	      }
	      if(argc > i+1 && argv[i+1][0] != '-') {
		version = argv[i+1];
		i++;
	      }
	      printf("format: %s version: %s\n", format, version);
	      err = cmdNew(ppSigDoc, format, version);
	      RETURN_IF_NOT(err == ERR_OK, err);
	    }
	    // add a DataFile
	    if(!strcmp(argv[i], "-add")) {
	      if(argc > i+2 && argv[i+1][0] != '-' && argv[i+2][0] != '-') {
		char* file = argv[i+1];
		char* mime = argv[i+2];
		char* content = CONTENT_EMBEDDED_BASE64;
		char* charset = CHARSET_UTF_8;
		i += 2;
		// optional content and charset
		if(argc > i+1 && argv[i+1][0] != '-') {
		  content = argv[i+1];
		  i++;
		}
		if(argc > i+1 && argv[i+1][0] != '-') {
		  charset = argv[i+1];
		  i++;
		}
		err = cmdAddDataFile(ppSigDoc, (const char*)file, (const char*)mime,
				     (const char*)content, (const char*)charset);
	      } else {
		printf("Missing <file> and <mime-type> arguments of -add command\n");	
	      }
	    }
	    
	  }
	  if(err)
	    printUsage();
  	}
	return err;
}

//--------------------------------------------------
// Add ecryption recipients
//--------------------------------------------------
int runRecipientCmds(int argc, char** argv, DEncEncryptedData** ppEncData)
{
  int err = ERR_OK, i;
  DEncEncryptedKey* pEncKey = 0;
  X509* pCert = 0;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // add a recipients key
      if(!strcmp(argv[i], "-encrecv")) {
	if(argc > i+1 && argv[i+1][0] != '-') {
	  char* certfile = argv[i+1];
	  char* recipient = NULL;
	  char* keyname = NULL;
	  char* carriedkeyname = NULL;
	  char* id = NULL;
	  i++;
	  if(!(*ppEncData))
	    err = dencEncryptedData_new(ppEncData, DENC_XMLNS_XMLENC, DENC_ENC_METHOD_AES128, 0, 0, 0);
	  // optional arguments
	  checkArguments(argc, argv, &i, &recipient);
	  checkArguments(argc, argv, &i, &keyname);
	  checkArguments(argc, argv, &i, &carriedkeyname);
	  err = ReadCertificate(&pCert, certfile);
	  if(err) return err;
	  err = dencEncryptedKey_new(*ppEncData, &pEncKey, pCert,
				     DENC_ENC_METHOD_RSA1_5,
				     id, recipient, keyname, carriedkeyname);
	} else {
	  printf("Missing <cert-file> argument of -encrecv command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}

//--------------------------------------------------
// Encrypts whole files
//--------------------------------------------------
int runEncryptFileCmds(int argc, char** argv, DEncEncryptedData* pEncData)
{
  int err = ERR_OK, i;
  DEncEncryptedKey* pEncKey = 0;
  X509* pCert = 0;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // add a recipients key
      if(!strcmp(argv[i], "-encrypt-file")) {
	if(argc > i+2 && argv[i+1][0] != '-' && argv[i+2][0] != '-') {
	  char* infile = NULL;
	  char* outfile = NULL;
	  char* mime = NULL;
	  convertInput((const char*)argv[i+1], &infile);
	  convertInput((const char*)argv[i+2], &outfile);
	  i += 2;
	  // optional arguments
	  checkArguments(argc, argv, &i, &mime);
	  // encrypt the file
	  err = dencEncryptFile(pEncData, infile, outfile, mime);
	  if(err) return err;
	} else {
	  printf("Missing <in-file> or <out-file> argument of -encrypt-file command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}


//--------------------------------------------------
// Decrypts whole files
//--------------------------------------------------
int runDecryptFileCmds(int argc, char** argv)
{
  int err = ERR_OK, i;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // decrypt a file
      if(!strcmp(argv[i], "-decrypt-file")) {
	if(argc > i+2 && argv[i+1][0] != '-' && argv[i+2][0] != '-' && 
	   argv[i+3][0] != '-' && argv[i+4][0] != '-') {
	  char* infile = NULL;
	  char* outfile = NULL;
	  char* cert_cn = NULL;
	  char* pin = NULL;
	  convertInput((const char*)argv[i+1], &infile);
	  convertInput((const char*)argv[i+2], &outfile);
	  convertInput((const char*)argv[i+3], &cert_cn);
	  convertInput((const char*)argv[i+4], &pin);
	  i += 4;
	  // decrypt the file
	  err = dencSaxReadDecryptFile(infile, outfile, cert_cn, pin);
	  if(err) return err;
	} else {
	  printf("Missing <in-file>, <out-file>, <cert-cn> or <pin> argument of -decrypt-file command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}


//--------------------------------------------------
// Add ecryption recipients
//--------------------------------------------------
int runDEncListCmds(int argc, char** argv)
{
  int err = ERR_OK, i, k;
  DEncEncryptedData* pEncData = 0;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // add a recipients key
      if(!strcmp(argv[i], "-denc-list")) {
	if(argc > i+1 && argv[i+1][0] != '-') {
	  char* file = argv[i+1];
	  err = dencSaxReadEncryptedData(&pEncData, file);
	  if(!err) {
	    printf("EncryptedData - data-status: %d data-len: %d key-status: %d key-len: %d\n", 
		   pEncData->nDataStatus, pEncData->mbufEncryptedData.nLen, 
		   pEncData->nKeyStatus, pEncData->mbufTransportKey.nLen);
	    if(pEncData->szId) printf("  Id=%s\n", pEncData->szId);
	    if(pEncData->szType) printf("  Type=%s\n", pEncData->szType);
	    if(pEncData->szMimeType) printf("  MimeType=%s\n", pEncData->szMimeType);
	    if(pEncData->szEncryptionMethod) printf("  EncryptionMethod=%s\n", pEncData->szEncryptionMethod);
	    for(k = 0; k < pEncData->nEncryptedKeys; k++) {
	      DEncEncryptedKey *pEncKey = pEncData->arrEncryptedKeys[k];
	      printf("\tEncryptedKey: key-len: %d Cert: %s\n",
		     pEncKey->mbufTransportKey.nLen, (pEncKey->pCert ? "OK" : "NULL"));
	      if(pEncKey->szId) printf("\t  Id=%s\n", pEncKey->szId);
	      if(pEncKey->szRecipient) printf("\t  Recipient=%s\n", pEncKey->szRecipient);
	      if(pEncKey->szKeyName) printf("\t  KeyName=%s\n", pEncKey->szKeyName);
	      if(pEncKey->szCarriedKeyName) printf("\t  CarriedKeyName=%s\n", pEncKey->szCarriedKeyName);
	      if(pEncKey->szEncryptionMethod) printf("\t  EncryptionMethod=%s\n", pEncKey->szEncryptionMethod);
	    }
	    if(pEncData->encProperties.nEncryptionProperties > 0) {
	      printf("\tEncryptionProperties: %s\n", 
		     (pEncData->encProperties.szId ? pEncData->encProperties.szId : ""));
	      for(k = 0; k < pEncData->encProperties.nEncryptionProperties; k++) {
		DEncEncryptionProperty *pEncProp = pEncData->encProperties.arrEncryptionProperties[k];
		printf("\t   Property\n");
		if(pEncProp->szId) printf("\t\tId=%s\n", pEncProp->szId);
		if(pEncProp->szTarget) printf("\t\tTarget=%s\n", pEncProp->szTarget);
		if(pEncProp->szName) printf("\t\tName=%s\n", pEncProp->szName);
		if(pEncProp->szContent) printf("\t\tContent=%s\n", pEncProp->szContent);
	      }
	    }

	  } else
	    printf("Error: %d reading file: %s\n", err, file);
	  if(pEncData)
	    dencEncryptedData_free(pEncData);
	} else {
	  printf("Missing <file> argument of -denc-list command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}

//--------------------------------------------------
// decryption file 
//--------------------------------------------------
int runDecryptCmds(int argc, char** argv, DEncEncryptedData** ppEncData)
{
  int err = ERR_OK, i;
  DEncEncryptedKey* pEncKey = 0;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // decrypt file
      if(!strcmp(argv[i], "-decrypt")) {
	if(argc > i+3 && 
	   argv[i+1][0] != '-' && 
	   argv[i+2][0] != '-' && 
	   argv[i+3][0] != '-') {
	  char* infile = argv[i+1];
	  char* recipient = argv[i+2];
	  char* pin = argv[i+3];
	  i += 3;
	  err = dencSaxReadEncryptedData(ppEncData, infile);
	  if(err) return err;
	  pEncKey = dencEncryptedData_FindEncryptedKeyByRecipient(*ppEncData, recipient);
	  if(pEncKey) {
	    err = dencEncryptedData_decrypt(*ppEncData, pEncKey, pin);
	    if(err) return err;
	  } else
	    printf("No such recipient: \'%s\'\n", recipient);
	  // p_szOutFile
	} else {
	  printf("Missing <file>, <recipient> or <pin> argument of -decrypt command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}

//--------------------------------------------------
// Create digidoc and adds datafiles
//--------------------------------------------------
int runExtractCmds(int argc, char** argv, SignedDoc** ppSigDoc)
{
  int err = ERR_OK, i;

  for(i = 1; (err == ERR_OK) && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // add a DataFile
      if(!strcmp(argv[i], "-extract")) {
	if(argc > i+2 && argv[i+1][0] != '-' && argv[i+2][0] != '-') {
	  char* id = NULL;
	  char* file = NULL;
	  char* charset = CHARSET_UTF_8;
	  checkArguments(argc, argv, &i, &id);
	  checkArguments(argc, argv, &i, &file);
	  // optional charset & filenamecharset
	  //checkArguments(argc, argv, &i, &charset);
	  err = ddocSaxExtractDataFile(*ppSigDoc, (const char*)p_szInFile, (const char*)file,
	    (const char*)id, (const char*)charset);
	  /*err = ddocXRdrExtractDataFile(*ppSigDoc, (const char*)p_szInFile, (const char*)file,
				(const char*)id, (const char*)charset, 
				(const char*)filenamecharset);*/
	  if(id) free(id);
	  if(file) free(file);
	  //if(charset) free(charset);

	} else {
	  printf("Missing <file> argument of -extract command\n");	
	}
      }
    }
    if(err)
      printUsage();
  }
  return err;
}

//--------------------------------------------------
// Verfys sigantures and notaries
//--------------------------------------------------
int runVerifyCmds(int argc, char** argv, SignedDoc** ppSigDoc)
{
	int err = ERR_OK, i;

	for(i = 1; !err && (i < argc); i++) {
		if(argv[i][0] == '-') { // all commands and options must start with -
			// create new digidoc
		      	if(!strcmp(argv[i], "-verify")) {
				err = cmdVerify(*ppSigDoc);
      			}
    		}
  	}
	return err;	
}


//--------------------------------------------------
// Handles signature commands
//--------------------------------------------------
int runSignCmds(int argc, char** argv, SignedDoc** ppSigDoc)
{
  int err = ERR_OK, i, nManif;

  for(i = 1; !err && (i < argc); i++) {
    if(argv[i][0] == '-') { // all commands and options must start with -
      // create new digidoc
      if(!strcmp(argv[i], "-sign")) {
	if(argc > i+1 && argv[i+1][0] != '-') {
	  char* pin = argv[i+1];
	  char* manifest = NULL;
	  char* city = NULL;
	  char* state = NULL;
	  char* zip = NULL;
	  char* country = NULL;
	  // check config file
	  nManif = ConfigItem_lookup_int("MANIFEST_MODE", 0);
	  if(nManif == 2) {
	    manifest = (char*)ConfigItem_lookup("DIGIDOC_ROLE_MANIFEST");
	    country = (char*)ConfigItem_lookup("DIGIDOC_ADR_COUNTRY");
	    state = (char*)ConfigItem_lookup("DIGIDOC_ADR_STATE");
	    city = (char*)ConfigItem_lookup("DIGIDOC_ADR_CITY");
	    zip = (char*)ConfigItem_lookup("DIGIDOC_ADR_ZIP");
	  }
	  i += 1;
	  // optional mainfest argument
	  checkArguments(argc, argv, &i, &manifest);
	  // optional address
	  checkArguments(argc, argv, &i, &city);
	  checkArguments(argc, argv, &i, &state);
	  checkArguments(argc, argv, &i, &zip);
	  checkArguments(argc, argv, &i, &country);	  
	  err = cmdSign(*ppSigDoc, (const char*)pin, (const char*)manifest,
			(const char*)city, (const char*)state, (const char*)zip, 
			(const char*)country);
	  if(manifest) free(manifest);
	  if(city) free(city);
	  if(state) free(state);
	  if(zip) free(zip);
	  if(country) free(country);
	} else {	
	  printf("Missing <pin> argument of -sign command\n");	
	}
      }
    }
  }
  return err;
}


//--------------------------------------------------
// Program main function
//--------------------------------------------------
int main(int argc, char** argv)
{
  int err = ERR_OK;
  SignedDoc* pSigDoc = 0;
  DEncEncryptedData* pEncData = 0;

  // display programm name and version
  printf("%s - %s\n", g_szProgName, g_szProgVer);
  if((err = checkProgArguments(argc, argv)) != ERR_OK) {
    checkDigiDocErrors();
    exit(1);
  }

  // init DigiDoc library
  initDigiDocLib();

  // read in config file
  err = initConfigStore();
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // use command line argument if required
  if(p_szConfigFile)
    err = readConfigFile(p_szConfigFile, ITEM_TYPE_PRIVATE);
  // execute the commands now 
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // read input file if necessary
  if(p_szInFile)
    err = cmdReadDigiDoc(&pSigDoc, p_parseMode);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  if(p_szInEncFile)
    err = cmdEncrypt(&pEncData, p_szInEncFile);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // list encrypted files
  err = runDEncListCmds(argc, argv);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // add data files
  err = runAddCmds(argc, argv, &pSigDoc);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // add recipients
  err = runRecipientCmds(argc, argv, &pEncData);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // encrypt whole files
  err = runEncryptFileCmds(argc, argv, pEncData);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // decrypt whole files
  err = runDecryptFileCmds(argc, argv);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // run signature commands
  err = runSignCmds(argc, argv, &pSigDoc);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // verify signatures
  err = runVerifyCmds(argc, argv, &pSigDoc);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // extract datafiles
  err = runExtractCmds(argc, argv, &pSigDoc);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // decrypt files
  err = runDecryptCmds(argc, argv, &pEncData);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}
  // write output file
  if(p_szOutFile)
    err = cmdWrite(pSigDoc, pEncData);
  if(err != ERR_OK) { checkDigiDocErrors(); exit(1);}  
  
  if(pSigDoc)
    SignedDoc_free(pSigDoc);
  if(pEncData)
    dencEncryptedData_free(pEncData);
  // cleanup
  cleanupConfigStore(NULL);
  finalizeDigiDocLib();

  return err;
}
